@@ -16,7 +16,6 @@ use std::env;
 
 use heck::{ToKebabCase, ToLowerCamelCase, ToShoutySnakeCase, ToSnakeCase, ToUpperCamelCase};
 use proc_macro2::{self, Span, TokenStream};
-use proc_macro_error::abort;
 use quote::{format_ident, quote, quote_spanned, ToTokens};
 use syn::DeriveInput;
 use syn::{self, ext::IdentExt, spanned::Spanned, Attribute, Field, Ident, LitStr, Type, Variant};
@@ -53,7 +52,7 @@ pub struct Item {
 }
 
 impl Item {
-    pub fn from_args_struct(input: &DeriveInput, name: Name) -> Self {
+    pub fn from_args_struct(input: &DeriveInput, name: Name) -> Result<Self, syn::Error> {
         let ident = input.ident.clone();
         let span = input.ident.span();
         let attrs = &input.attrs;
@@ -62,15 +61,15 @@ impl Item {
         let kind = Sp::new(Kind::Command(Sp::new(Ty::Other, span)), span);
 
         let mut res = Self::new(name, ident, None, argument_casing, env_casing, kind);
-        let parsed_attrs = ClapAttr::parse_all(attrs);
-        res.infer_kind(&parsed_attrs);
-        res.push_attrs(&parsed_attrs);
+        let parsed_attrs = ClapAttr::parse_all(attrs)?;
+        res.infer_kind(&parsed_attrs)?;
+        res.push_attrs(&parsed_attrs)?;
         res.push_doc_comment(attrs, "about", Some("long_about"));
 
-        res
+        Ok(res)
     }
 
-    pub fn from_subcommand_enum(input: &DeriveInput, name: Name) -> Self {
+    pub fn from_subcommand_enum(input: &DeriveInput, name: Name) -> Result<Self, syn::Error> {
         let ident = input.ident.clone();
         let span = input.ident.span();
         let attrs = &input.attrs;
@@ -79,15 +78,15 @@ impl Item {
         let kind = Sp::new(Kind::Command(Sp::new(Ty::Other, span)), span);
 
         let mut res = Self::new(name, ident, None, argument_casing, env_casing, kind);
-        let parsed_attrs = ClapAttr::parse_all(attrs);
-        res.infer_kind(&parsed_attrs);
-        res.push_attrs(&parsed_attrs);
+        let parsed_attrs = ClapAttr::parse_all(attrs)?;
+        res.infer_kind(&parsed_attrs)?;
+        res.push_attrs(&parsed_attrs)?;
         res.push_doc_comment(attrs, "about", Some("long_about"));
 
-        res
+        Ok(res)
     }
 
-    pub fn from_value_enum(input: &DeriveInput, name: Name) -> Self {
+    pub fn from_value_enum(input: &DeriveInput, name: Name) -> Result<Self, syn::Error> {
         let ident = input.ident.clone();
         let span = input.ident.span();
         let attrs = &input.attrs;
@@ -96,9 +95,9 @@ impl Item {
         let kind = Sp::new(Kind::Value, span);
 
         let mut res = Self::new(name, ident, None, argument_casing, env_casing, kind);
-        let parsed_attrs = ClapAttr::parse_all(attrs);
-        res.infer_kind(&parsed_attrs);
-        res.push_attrs(&parsed_attrs);
+        let parsed_attrs = ClapAttr::parse_all(attrs)?;
+        res.infer_kind(&parsed_attrs)?;
+        res.push_attrs(&parsed_attrs)?;
         // Ignoring `push_doc_comment` as there is no top-level clap builder to add documentation
         // to
 
@@ -110,14 +109,14 @@ impl Item {
             );
         }
 
-        res
+        Ok(res)
     }
 
     pub fn from_subcommand_variant(
         variant: &Variant,
         struct_casing: Sp<CasingStyle>,
         env_casing: Sp<CasingStyle>,
-    ) -> Self {
+    ) -> Result<Self, syn::Error> {
         let name = variant.ident.clone();
         let ident = variant.ident.clone();
         let span = variant.span();
@@ -138,9 +137,9 @@ impl Item {
             env_casing,
             kind,
         );
-        let parsed_attrs = ClapAttr::parse_all(&variant.attrs);
-        res.infer_kind(&parsed_attrs);
-        res.push_attrs(&parsed_attrs);
+        let parsed_attrs = ClapAttr::parse_all(&variant.attrs)?;
+        res.infer_kind(&parsed_attrs)?;
+        res.push_attrs(&parsed_attrs)?;
         if matches!(&*res.kind, Kind::Command(_) | Kind::Subcommand(_)) {
             res.push_doc_comment(&variant.attrs, "about", Some("long_about"));
         }
@@ -164,14 +163,14 @@ impl Item {
             | Kind::Arg(_) => (),
         }
 
-        res
+        Ok(res)
     }
 
     pub fn from_value_enum_variant(
         variant: &Variant,
         argument_casing: Sp<CasingStyle>,
         env_casing: Sp<CasingStyle>,
-    ) -> Self {
+    ) -> Result<Self, syn::Error> {
         let ident = variant.ident.clone();
         let span = variant.span();
         let kind = Sp::new(Kind::Value, span);
@@ -183,21 +182,21 @@ impl Item {
             env_casing,
             kind,
         );
-        let parsed_attrs = ClapAttr::parse_all(&variant.attrs);
-        res.infer_kind(&parsed_attrs);
-        res.push_attrs(&parsed_attrs);
+        let parsed_attrs = ClapAttr::parse_all(&variant.attrs)?;
+        res.infer_kind(&parsed_attrs)?;
+        res.push_attrs(&parsed_attrs)?;
         if matches!(&*res.kind, Kind::Value) {
             res.push_doc_comment(&variant.attrs, "help", None);
         }
 
-        res
+        Ok(res)
     }
 
     pub fn from_args_field(
         field: &Field,
         struct_casing: Sp<CasingStyle>,
         env_casing: Sp<CasingStyle>,
-    ) -> Self {
+    ) -> Result<Self, syn::Error> {
         let name = field.ident.clone().unwrap();
         let ident = field.ident.clone().unwrap();
         let span = field.span();
@@ -211,9 +210,9 @@ impl Item {
             env_casing,
             kind,
         );
-        let parsed_attrs = ClapAttr::parse_all(&field.attrs);
-        res.infer_kind(&parsed_attrs);
-        res.push_attrs(&parsed_attrs);
+        let parsed_attrs = ClapAttr::parse_all(&field.attrs)?;
+        res.infer_kind(&parsed_attrs)?;
+        res.push_attrs(&parsed_attrs)?;
         if matches!(&*res.kind, Kind::Arg(_)) {
             res.push_doc_comment(&field.attrs, "help", Some("long_help"));
         }
@@ -244,7 +243,7 @@ impl Item {
             | Kind::ExternalSubcommand => {}
         }
 
-        res
+        Ok(res)
     }
 
     fn new(
@@ -329,7 +328,7 @@ impl Item {
         }
     }
 
-    fn infer_kind(&mut self, attrs: &[ClapAttr]) {
+    fn infer_kind(&mut self, attrs: &[ClapAttr]) -> Result<(), syn::Error> {
         for attr in attrs {
             if let Some(AttrValue::Call(_)) = &attr.value {
                 continue;
@@ -339,7 +338,7 @@ impl Item {
             let kind = match &attr.magic {
                 Some(MagicAttrName::FromGlobal) => {
                     if attr.value.is_some() {
-                        let expr = attr.value_or_abort();
+                        let expr = attr.value_or_abort()?;
                         abort!(expr, "attribute `{}` does not accept a value", attr.name);
                     }
                     let ty = self
@@ -352,7 +351,7 @@ impl Item {
                 }
                 Some(MagicAttrName::Subcommand) if attr.value.is_none() => {
                     if attr.value.is_some() {
-                        let expr = attr.value_or_abort();
+                        let expr = attr.value_or_abort()?;
                         abort!(expr, "attribute `{}` does not accept a value", attr.name);
                     }
                     let ty = self
@@ -365,7 +364,7 @@ impl Item {
                 }
                 Some(MagicAttrName::ExternalSubcommand) if attr.value.is_none() => {
                     if attr.value.is_some() {
-                        let expr = attr.value_or_abort();
+                        let expr = attr.value_or_abort()?;
                         abort!(expr, "attribute `{}` does not accept a value", attr.name);
                     }
                     let kind = Sp::new(Kind::ExternalSubcommand, attr.name.clone().span());
@@ -373,7 +372,7 @@ impl Item {
                 }
                 Some(MagicAttrName::Flatten) if attr.value.is_none() => {
                     if attr.value.is_some() {
-                        let expr = attr.value_or_abort();
+                        let expr = attr.value_or_abort()?;
                         abort!(expr, "attribute `{}` does not accept a value", attr.name);
                     }
                     let ty = self
@@ -396,12 +395,14 @@ impl Item {
             };
 
             if let Some(kind) = kind {
-                self.set_kind(kind);
+                self.set_kind(kind)?;
             }
         }
+
+        Ok(())
     }
 
-    fn push_attrs(&mut self, attrs: &[ClapAttr]) {
+    fn push_attrs(&mut self, attrs: &[ClapAttr]) -> Result<(), syn::Error> {
         for attr in attrs {
             let actual_attr_kind = *attr.kind.get();
             let expected_attr_kind = self.kind.attr_kind();
@@ -437,7 +438,7 @@ impl Item {
 
             match &attr.magic {
                 Some(MagicAttrName::Short) if attr.value.is_none() => {
-                    assert_attr_kind(attr, &[AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Arg])?;
 
                     self.push_method(
                         *attr.kind.get(),
@@ -447,13 +448,13 @@ impl Item {
                 }
 
                 Some(MagicAttrName::Long) if attr.value.is_none() => {
-                    assert_attr_kind(attr, &[AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Arg])?;
 
                     self.push_method(*attr.kind.get(), attr.name.clone(), self.name.clone().translate(*self.casing));
                 }
 
                 Some(MagicAttrName::ValueParser) if attr.value.is_none() => {
-                    assert_attr_kind(attr, &[AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Arg])?;
 
                     self.deprecations.push(Deprecation {
                         span: attr.name.span(),
@@ -465,7 +466,7 @@ impl Item {
                 }
 
                 Some(MagicAttrName::Action) if attr.value.is_none() => {
-                    assert_attr_kind(attr, &[AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Arg])?;
 
                     self.deprecations.push(Deprecation {
                         span: attr.name.span(),
@@ -477,7 +478,7 @@ impl Item {
                 }
 
                 Some(MagicAttrName::Env) if attr.value.is_none() => {
-                    assert_attr_kind(attr, &[AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Arg])?;
 
                     self.push_method(
                         *attr.kind.get(),
@@ -487,7 +488,7 @@ impl Item {
                 }
 
                 Some(MagicAttrName::ValueEnum) if attr.value.is_none() => {
-                    assert_attr_kind(attr, &[AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Arg])?;
 
                     self.is_enum = true
                 }
@@ -497,45 +498,45 @@ impl Item {
                 }
 
                 Some(MagicAttrName::About) if attr.value.is_none() => {
-                    assert_attr_kind(attr, &[AttrKind::Command]);
+                    assert_attr_kind(attr, &[AttrKind::Command])?;
 
                     if let Some(method) =
-                        Method::from_env(attr.name.clone(), "CARGO_PKG_DESCRIPTION")
+                        Method::from_env(attr.name.clone(), "CARGO_PKG_DESCRIPTION")?
                     {
                         self.methods.push(method);
                     }
                 }
 
                 Some(MagicAttrName::LongAbout) if attr.value.is_none() => {
-                    assert_attr_kind(attr, &[AttrKind::Command]);
+                    assert_attr_kind(attr, &[AttrKind::Command])?;
 
                     self.force_long_help = true;
                 }
 
                 Some(MagicAttrName::LongHelp) if attr.value.is_none() => {
-                    assert_attr_kind(attr, &[AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Arg])?;
 
                     self.force_long_help = true;
                 }
 
                 Some(MagicAttrName::Author) if attr.value.is_none() => {
-                    assert_attr_kind(attr, &[AttrKind::Command]);
+                    assert_attr_kind(attr, &[AttrKind::Command])?;
 
-                    if let Some(method) = Method::from_env(attr.name.clone(), "CARGO_PKG_AUTHORS") {
+                    if let Some(method) = Method::from_env(attr.name.clone(), "CARGO_PKG_AUTHORS")? {
                         self.methods.push(method);
                     }
                 }
 
                 Some(MagicAttrName::Version) if attr.value.is_none() => {
-                    assert_attr_kind(attr, &[AttrKind::Command]);
+                    assert_attr_kind(attr, &[AttrKind::Command])?;
 
-                    if let Some(method) = Method::from_env(attr.name.clone(), "CARGO_PKG_VERSION") {
+                    if let Some(method) = Method::from_env(attr.name.clone(), "CARGO_PKG_VERSION")? {
                         self.methods.push(method);
                     }
                 }
 
                 Some(MagicAttrName::DefaultValueT) => {
-                    assert_attr_kind(attr, &[AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Arg])?;
 
                     let ty = if let Some(ty) = self.ty.as_ref() {
                         ty
@@ -543,7 +544,7 @@ impl Item {
                         abort!(
                             attr.name.clone(),
                             "#[arg(default_value_t)] (without an argument) can be used \
-                            only on field level";
+                            only on field level\n\n= note: {note}\n\n",
 
                             note = "see \
                                 https://github.com/clap-rs/clap/blob/master/examples/derive_ref/README.md#magic-attributes")
@@ -583,7 +584,7 @@ impl Item {
                 }
 
                 Some(MagicAttrName::DefaultValuesT) => {
-                    assert_attr_kind(attr, &[AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Arg])?;
 
                     let ty = if let Some(ty) = self.ty.as_ref() {
                         ty
@@ -591,18 +592,18 @@ impl Item {
                         abort!(
                             attr.name.clone(),
                             "#[arg(default_values_t)] (without an argument) can be used \
-                            only on field level";
+                            only on field level\n\n= note: {note}\n\n",
 
                             note = "see \
                                 https://github.com/clap-rs/clap/blob/master/examples/derive_ref/README.md#magic-attributes")
                     };
-                    let expr = attr.value_or_abort();
+                    let expr = attr.value_or_abort()?;
 
                     let container_type = Ty::from_syn_ty(ty);
                     if *container_type != Ty::Vec {
                         abort!(
                             attr.name.clone(),
-                            "#[arg(default_values_t)] can be used only on Vec types";
+                            "#[arg(default_values_t)] can be used only on Vec types\n\n= note: {note}\n\n",
 
                             note = "see \
                                 https://github.com/clap-rs/clap/blob/master/examples/derive_ref/README.md#magic-attributes")
@@ -667,7 +668,7 @@ impl Item {
                 }
 
                 Some(MagicAttrName::DefaultValueOsT) => {
-                    assert_attr_kind(attr, &[AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Arg])?;
 
                     let ty = if let Some(ty) = self.ty.as_ref() {
                         ty
@@ -675,7 +676,7 @@ impl Item {
                         abort!(
                             attr.name.clone(),
                             "#[arg(default_value_os_t)] (without an argument) can be used \
-                            only on field level";
+                            only on field level\n\n= note: {note}\n\n",
 
                             note = "see \
                                 https://github.com/clap-rs/clap/blob/master/examples/derive_ref/README.md#magic-attributes")
@@ -715,7 +716,7 @@ impl Item {
                 }
 
                 Some(MagicAttrName::DefaultValuesOsT) => {
-                    assert_attr_kind(attr, &[AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Arg])?;
 
                     let ty = if let Some(ty) = self.ty.as_ref() {
                         ty
@@ -723,18 +724,18 @@ impl Item {
                         abort!(
                             attr.name.clone(),
                             "#[arg(default_values_os_t)] (without an argument) can be used \
-                            only on field level";
+                            only on field level\n\n= note: {note}\n\n",
 
                             note = "see \
                                 https://github.com/clap-rs/clap/blob/master/examples/derive_ref/README.md#magic-attributes")
                     };
-                    let expr = attr.value_or_abort();
+                    let expr = attr.value_or_abort()?;
 
                     let container_type = Ty::from_syn_ty(ty);
                     if *container_type != Ty::Vec {
                         abort!(
                             attr.name.clone(),
-                            "#[arg(default_values_os_t)] can be used only on Vec types";
+                            "#[arg(default_values_os_t)] can be used only on Vec types\n\n= note: {note}\n\n",
 
                             note = "see \
                                 https://github.com/clap-rs/clap/blob/master/examples/derive_ref/README.md#magic-attributes")
@@ -799,29 +800,29 @@ impl Item {
                 }
 
                 Some(MagicAttrName::NextDisplayOrder) => {
-                    assert_attr_kind(attr, &[AttrKind::Command]);
+                    assert_attr_kind(attr, &[AttrKind::Command])?;
 
-                    let expr = attr.value_or_abort();
+                    let expr = attr.value_or_abort()?;
                     self.next_display_order = Some(Method::new(attr.name.clone(), quote!(#expr)));
                 }
 
                 Some(MagicAttrName::NextHelpHeading) => {
-                    assert_attr_kind(attr, &[AttrKind::Command]);
+                    assert_attr_kind(attr, &[AttrKind::Command])?;
 
-                    let expr = attr.value_or_abort();
+                    let expr = attr.value_or_abort()?;
                     self.next_help_heading = Some(Method::new(attr.name.clone(), quote!(#expr)));
                 }
 
                 Some(MagicAttrName::RenameAll) => {
-                    let lit = attr.lit_str_or_abort();
-                    self.casing = CasingStyle::from_lit(lit);
+                    let lit = attr.lit_str_or_abort()?;
+                    self.casing = CasingStyle::from_lit(lit)?;
                 }
 
                 Some(MagicAttrName::RenameAllEnv) => {
-                    assert_attr_kind(attr, &[AttrKind::Command, AttrKind::Arg]);
+                    assert_attr_kind(attr, &[AttrKind::Command, AttrKind::Arg])?;
 
-                    let lit = attr.lit_str_or_abort();
-                    self.env_casing = CasingStyle::from_lit(lit);
+                    let lit = attr.lit_str_or_abort()?;
+                    self.env_casing = CasingStyle::from_lit(lit)?;
                 }
 
                 Some(MagicAttrName::Skip) if actual_attr_kind == AttrKind::Group => {
@@ -839,20 +840,20 @@ impl Item {
                 | Some(MagicAttrName::Author)
                 | Some(MagicAttrName::Version)
                  => {
-                    let expr = attr.value_or_abort();
+                    let expr = attr.value_or_abort()?;
                     self.push_method(*attr.kind.get(), attr.name.clone(), expr);
                 }
 
                 // Magic only for the default, otherwise just forward to the builder
                 Some(MagicAttrName::ValueParser) | Some(MagicAttrName::Action) => {
-                    let expr = attr.value_or_abort();
+                    let expr = attr.value_or_abort()?;
                     self.push_method(*attr.kind.get(), attr.name.clone(), expr);
                 }
 
                 // Directives that never receive a value
                 Some(MagicAttrName::ValueEnum)
                 | Some(MagicAttrName::VerbatimDocComment) => {
-                    let expr = attr.value_or_abort();
+                    let expr = attr.value_or_abort()?;
                     abort!(expr, "attribute `{}` does not accept a value", attr.name);
                 }
 
@@ -883,6 +884,8 @@ impl Item {
                 );
             }
         }
+
+        Ok(())
     }
 
     fn push_doc_comment(&mut self, attrs: &[Attribute], short_name: &str, long_name: Option<&str>) {
@@ -912,7 +915,7 @@ impl Item {
         }
     }
 
-    fn set_kind(&mut self, kind: Sp<Kind>) {
+    fn set_kind(&mut self, kind: Sp<Kind>) -> Result<(), syn::Error> {
         match (self.kind.get(), kind.get()) {
             (Kind::Arg(_), Kind::FromGlobal(_))
             | (Kind::Arg(_), Kind::Subcommand(_))
@@ -932,6 +935,7 @@ impl Item {
                 abort!(kind.span(), "`{}` cannot be used with `{}`", new, old);
             }
         }
+        Ok(())
     }
 
     pub fn find_default_method(&self) -> Option<&Method> {
@@ -1121,7 +1125,7 @@ impl Action {
 fn default_action(field_type: &Type, span: Span) -> Method {
     let ty = Ty::from_syn_ty(field_type);
     let args = match *ty {
-        Ty::Vec | Ty::OptionVec => {
+        Ty::Vec | Ty::OptionVec | Ty::VecVec | Ty::OptionVecVec => {
             quote_spanned! { span=>
                 clap::ArgAction::Append
             }
@@ -1211,19 +1215,21 @@ impl Method {
         Method { name, args }
     }
 
-    fn from_env(ident: Ident, env_var: &str) -> Option<Self> {
+    fn from_env(ident: Ident, env_var: &str) -> Result<Option<Self>, syn::Error> {
         let mut lit = match env::var(env_var) {
             Ok(val) => {
                 if val.is_empty() {
-                    return None;
+                    return Ok(None);
                 }
                 LitStr::new(&val, ident.span())
             }
             Err(_) => {
-                abort!(ident,
-                    "cannot derive `{}` from Cargo.toml", ident;
-                    note = "`{}` environment variable is not set", env_var;
-                    help = "use `{} = \"...\"` to set {} manually", ident, ident;
+                abort!(
+                    ident,
+                    "cannot derive `{}` from Cargo.toml\n\n= note: {note}\n\n= help: {help}\n\n",
+                    ident,
+                    note = format_args!("`{}` environment variable is not set", env_var),
+                    help = format_args!("use `{} = \"...\"` to set {} manually", ident, ident)
                 );
             }
         };
@@ -1233,7 +1239,7 @@ impl Method {
             lit = LitStr::new(&edited, lit.span());
         }
 
-        Some(Method::new(ident, quote!(#lit)))
+        Ok(Some(Method::new(ident, quote!(#lit))))
     }
 
     pub(crate) fn args(&self) -> &TokenStream {
@@ -1299,7 +1305,7 @@ impl ToTokens for Deprecation {
     }
 }
 
-fn assert_attr_kind(attr: &ClapAttr, possible_kind: &[AttrKind]) {
+fn assert_attr_kind(attr: &ClapAttr, possible_kind: &[AttrKind]) -> Result<(), syn::Error> {
     if *attr.kind.get() == AttrKind::Clap || *attr.kind.get() == AttrKind::StructOpt {
         // deprecated
     } else if !possible_kind.contains(attr.kind.get()) {
@@ -1315,6 +1321,7 @@ fn assert_attr_kind(attr: &ClapAttr, possible_kind: &[AttrKind]) {
             options.join(", ")
         );
     }
+    Ok(())
 }
 
 /// replace all `:` with `, ` when not inside the `<>`
@@ -1364,13 +1371,13 @@ pub enum CasingStyle {
 }
 
 impl CasingStyle {
-    fn from_lit(name: &LitStr) -> Sp<Self> {
+    fn from_lit(name: &LitStr) -> Result<Sp<Self>, syn::Error> {
         use self::CasingStyle::*;
 
         let normalized = name.value().to_upper_camel_case().to_lowercase();
         let cs = |kind| Sp::new(kind, name.span());
 
-        match normalized.as_ref() {
+        let s = match normalized.as_ref() {
             "camel" | "camelcase" => cs(Camel),
             "kebab" | "kebabcase" => cs(Kebab),
             "pascal" | "pascalcase" => cs(Pascal),
@@ -1380,7 +1387,8 @@ impl CasingStyle {
             "upper" | "uppercase" => cs(Upper),
             "verbatim" | "verbatimcase" => cs(Verbatim),
             s => abort!(name, "unsupported casing: `{}`", s),
-        }
+        };
+        Ok(s)
     }
 }
 
